/******************************************************************************
* idmap_cdc.c
*
* Copyright (C) David Daugherty, 2007
* All Rights Reserved
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*
* A Samba ID Mapper implementation for Centrify's Direct Control using
* Centrify's CAPI interface.
*
******************************************************************************/

#include "includes.h"
#include "centrifydc.h"

typedef struct
{
    void *dlhandle;
    cdcSession_t session;

    cdcCode (*cdcSessionOpen)(int majorVersion,
                              int minorVersion,
                              cdcSession_t *session);
    
    void (*cdcSessionClose)(cdcSession_t session);

    cdcCode (*cdcSetOption)(cdcSession_t session,
                            int option,
                            void *option_value,
                            size_t option_size);

    cdcCode (*cdcLookupObjectByUnixId)(cdcSession_t session,
                                       int objType,
                                       int unixId,
                                       cdcObject_t *result);
    
    cdcCode (*cdcLookupObjectBySid)(cdcSession_t session,
                                    const char *sid,
                                    cdcObject_t *result);
    
    cdcBool (*cdcIsObjectCategory)(cdcSession_t session,
                                   cdcObject_t obj,
                                   const char *objCategory);

    cdcCode (*cdcGetAttributeStr)(cdcSession_t session,
                               cdcObject_t obj,
                               const char *attrName,
                                  char **result);
    
    cdcCode (*cdcGetAttributeInt)(cdcSession_t session,
                               cdcObject_t obj,
                               const char *attrName,
                               int *result);

    void (*cdcFreeString)(char *string);

    void (*cdcFreeObject)(cdcObject_t obj);

    const char* (*cdcGetCdcCodeStr)(cdcCode cc);

    const char* (*cdcGetErrSystemStr)(cdcErrSystem errSystem);

    cdcCode (*cdcGetLastProtocolError)(cdcSession_t session,
                                       cdcErrSystem *errSystem,
                                       int *errCode,
                                       const char **errString);

} CapiFuncs;

#define CDC_EXEC(CMD)                   \
{                                       \
    ccode = cf->CMD;                    \
    if (idmap_cdc_reconnect(cf, ccode)) \
        ccode = cf->CMD;                \
}

typedef struct
{
    cdcCode cdcErrCode;
    NTSTATUS ntErrCode;
} CdcCodeNTStatusMap;

CdcCodeNTStatusMap cdcNTStatusMap[] =
{
    { CDC_CODE_SUCCESS,                     NT_STATUS_OK },
    { CDC_CODE_FAILURE,                     NT_STATUS_UNSUCCESSFUL },
    { CDC_CODE_NOMEM,                       NT_STATUS_NO_MEMORY },
    { CDC_CODE_BAD_OPTION,                  NT_STATUS_INVALID_PARAMETER },
    { CDC_CODE_BAD_PARAM,                   NT_STATUS_INVALID_PARAMETER },
    { CDC_CODE_BAD_SESSION,                 NT_STATUS_NOT_CLIENT_SESSION },
    { CDC_CODE_LRPC_FAILED,                 NT_STATUS_NOT_CLIENT_SESSION },
    { CDC_CODE_NO_MORE,                     NT_STATUS_NO_MORE_ENTRIES },
    { CDC_CODE_NO_SUCH_ATTR,                NT_STATUS_RESOURCE_NAME_NOT_FOUND },
    { CDC_CODE_NO_SUCH_OBJECT,              NT_STATUS_NONE_MAPPED },
    { CDC_CODE_SERVER_UNREACHABLE,          NT_STATUS_HOST_UNREACHABLE },
    { CDC_CODE_SEARCH_IN_PROGRESS,          NT_STATUS_MORE_PROCESSING_REQUIRED },
    { CDC_CODE_BAD_VERSION,                 NT_STATUS_UNKNOWN_REVISION },
    { CDC_CODE_INVALID_USER,                NT_STATUS_NO_SUCH_USER },
    { CDC_CODE_INVALID_PASSWORD,            NT_STATUS_WRONG_PASSWORD },
    { CDC_CODE_ACCOUNT_LOCKED,              NT_STATUS_ACCOUNT_LOCKED_OUT },
    { CDC_CODE_PASSWORD_EXPIRED,            NT_STATUS_PASSWORD_EXPIRED },
    { CDC_CODE_PASSWORD_POLICY_NOT_MATCHED, NT_STATUS_PASSWORD_RESTRICTION },
    { CDC_CODE_PASSWORD_CHANGE_REJECTED,    NT_STATUS_ILL_FORMED_PASSWORD },
    { CDC_CODE_ACCOUNT_EXPIRED,             NT_STATUS_ACCOUNT_EXPIRED },
    { CDC_CODE_ACCOUNT_DISABLED,            NT_STATUS_ACCOUNT_DISABLED },
    { CDC_CODE_WORKSTATION_DENIED,          NT_STATUS_INVALID_WORKSTATION },
    { CDC_CODE_PERMISSION,                  NT_STATUS_ACCESS_DENIED },
    { CDC_CODE_BAD_PACKET,                  NT_STATUS_DATA_ERROR },
    { CDC_CODE_BAD_DATA,                    NT_STATUS_DATA_ERROR },
    { CDC_CODE_NOT_JOINED,                  NT_STATUS_NO_TRUST_SAM_ACCOUNT },
    { CDC_CODE_VALUE_NOT_SET,               NT_STATUS_RESOURCE_DATA_NOT_FOUND },
    { CDC_CODE_IO_ERROR,                    NT_STATUS_UNEXPECTED_IO_ERROR },
    { CDC_CODE_SYS_ERROR,                   NT_STATUS_INTERNAL_ERROR },
    { CDC_CODE_NO_SYS_ERROR_INFO,           NT_STATUS_SECTION_NOT_EXTENDED },
    { CDC_CODE_WRONG_DATA_TYPE,             ERROR_INVALID_DATATYPE },
    { CDC_CODE_MULTI_VALUE,                 ERROR_INVALID_DATATYPE },
    { CDC_CODE_NO_ADCLIENT,                 NT_STATUS_NO_LOGON_SERVERS },
    { CDC_CODE_LOGON_FAILURE,               NT_STATUS_LOGON_FAILURE },
};

#define cdcNTStatusMapSize (sizeof(cdcNTStatusMap) / sizeof(*cdcNTStatusMap))

/**
 * Map an internal CAPI error code to an NTSTATUS code.
 * Samba wants NTStatus codes.
 *
 * @param ccode - CAPI return code.
 * 
 * @return NTSTATUS - Mapped NTSTATUS return code.
 */
static NTSTATUS ccode_2_ntstatus(cdcCode ccode)
{
    // Return this error if we do not have a mapping
    NTSTATUS ntstatus = NT_STATUS_UNSUCCESSFUL;
    int i;

    for (i=0; i<cdcNTStatusMapSize; i++)
    {
        if (ccode == cdcNTStatusMap[i].cdcErrCode)
        {
            ntstatus = cdcNTStatusMap[i].ntErrCode;
            break;
        }
    }

    DEBUG(10, ("ccode=%d mapped to ntstatus=0x%x\n",
               ccode, NT_STATUS_V(ntstatus)));

    return ntstatus;
}

/**
 * Log extended error information.
 */
static void idmap_cdc_err_info(CapiFuncs *cf, cdcCode ccode)
{
    cdcErrSystem errSystem;
    int errCode;
    const char *errString;

    if (ccode == CDC_CODE_BAD_SESSION || ccode == CDC_CODE_LRPC_FAILED)
    {
        DEBUG(5, ("Capi session error ccode=%d: %s.\n",
                  ccode,
                  cf->cdcGetCdcCodeStr(ccode)));
    }
    else if (cf->cdcGetLastProtocolError(cf->session,
                                         &errSystem,
                                         &errCode,
                                         &errString) == CDC_CODE_SUCCESS)
    {
        DEBUG(10, ("CAPI Protocol Error Info: errSystem=%s, errCode=%d (0x%x) errString=%s\n",
               cf->cdcGetErrSystemStr(errSystem),
               errCode,
               errCode,
               errString));

        cf->cdcFreeString((char*)errString);
    }
}

/**
 * For certain errors we will attempt to re-establish a connection with
 * adclient.
 *
 * @param cf    - (IN) CAPI function pointers
 * @param ccode - (IN) Return code from last command.
 * 
 * @return      1 if session was closed and successfully reopened.
 */
static int idmap_cdc_reconnect(CapiFuncs *cf, cdcCode ccode)
{
    int reconnected = 0;

    if (ccode == CDC_CODE_BAD_SESSION || ccode == CDC_CODE_LRPC_FAILED)
    {
        DEBUG(5, ("Session is closed. ccode=%d: %s. Attempting to reopen...\n",
                  ccode,
                  cf->cdcGetCdcCodeStr(ccode)));
        
        if (cf->session)
        {
            cf->cdcSessionClose(cf->session);
            cf->session = 0;
        }

        ccode = cf->cdcSessionOpen(CDC_CAPI_MAJOR_VERSION,
                                   CDC_CAPI_MINOR_VERSION,
                                   &cf->session);

        if (ccode != CDC_CODE_SUCCESS)
        {
            DEBUG(1, ("Reconnect failed. DirectControl not loaded? ccode=%d %s\n",
                      ccode,
                      cf->cdcGetCdcCodeStr(ccode)));
        }
        else
        {
            reconnected = 1;
        }
    }

    return reconnected;
}

static void idmap_cdc_obj_info(cdcCode ccode,
                               const char *method,
                               CapiFuncs *cf,
                               cdcObject_t obj,
                               struct id_map *idmap,
                               const char *sidstr)
{
    if (ccode == CDC_CODE_SUCCESS)
    {
        char *name = 0;

        cf->cdcGetAttributeStr(cf->session, obj, CDC_AD_ATTR_NAME, &name);

        DEBUG(10, ("Lookup by %s: id=%d type=%d sid=%s name=%s\n",
                  method,
                  idmap->xid.id,
                  idmap->xid.type,
                  sidstr,
                  name ? name : "<unknown>"));

        if (name)
        {
            cf->cdcFreeString(name);
        }
    }
}

/**
 * Dynamically load the CAPI shared object library
 *
 * @param libpath   - path to CAPI shared object library.
 *
 * @return NTSTATUS - NT_STATUS_OK if dynamic loading okay.
 *                  - NT_STATUS_NO_SUCH_FILE if CAPI lib not found.
 */
static NTSTATUS idmap_cdc_load_lib(struct idmap_domain *dom)
{
    NTSTATUS ntstatus = NT_STATUS_NO_SUCH_FILE;
    CapiFuncs *cf = (CapiFuncs*)dom->private_data;
    const char *libpath = dom->params;
    fstring config_option;
    
    if (libpath == 0)
    {
        snprintf(config_option, sizeof(config_option), "idmap config %s", dom->name);

        libpath = lp_parm_const_string(-1, config_option, "params", NULL);

        if (libpath == 0)
        {
            DEBUG(2, ("smb.conf missing CAPI library parameter!\n"));
            goto ExitRoutine;
        }
    }

    memset(cf, 0, sizeof(*cf));
#ifdef AIX	
    cf->dlhandle = dlopen(libpath, RTLD_LAZY  | RTLD_MEMBER);
#else
    cf->dlhandle = dlopen(libpath, RTLD_LAZY);
#endif
    if (cf->dlhandle == NULL)
    {
        DEBUG(2, ("Failed to dlopen CAPI library: %s. Check smb.conf settings.\n", libpath));
    }
    else
    {
        cf->cdcSessionOpen = (cdcCode (*)
                             (int majorVersion,
                              int minorVersion,
                              cdcSession_t *session)
                            ) dlsym(cf->dlhandle, "cdcSessionOpen");

        cf->cdcSessionClose = (
                              void (*)(cdcSession_t session)
                             ) dlsym(cf->dlhandle, "cdcSessionClose");

        cf->cdcSetOption = (
                            cdcCode (*)(cdcSession_t session,
                                        int option,
                                        void *option_value,
                                        size_t option_size)
                           ) dlsym(cf->dlhandle, "cdcSetOption");

        cf->cdcLookupObjectByUnixId = (
                                      cdcCode (*)(cdcSession_t session,
                                        int objType,
                                        int unixId,
                                        cdcObject_t *result)
                                     ) dlsym(cf->dlhandle, "cdcLookupObjectByUnixId");

        cf->cdcLookupObjectBySid = (cdcCode (*)
                                   (cdcSession_t session,
                                    const char *sid,
                                    cdcObject_t *result)
                                  ) dlsym(cf->dlhandle, "cdcLookupObjectBySid");

        cf->cdcIsObjectCategory = (cdcBool (*)
                                  (cdcSession_t session,
                                   cdcObject_t obj,
                                   const char *objCategory)
                                 ) dlsym(cf->dlhandle, "cdcIsObjectCategory");

        cf->cdcGetAttributeStr = (cdcCode (*)
                                 (cdcSession_t session,
                                  cdcObject_t obj,
                                  const char *attrName,
                                  char **result)
                                ) dlsym(cf->dlhandle, "cdcGetAttributeStr");

        cf->cdcGetAttributeInt = (cdcCode (*)
                                 (cdcSession_t session,
                                  cdcObject_t obj,
                                  const char *attrName,
                                  int *result)
                                ) dlsym(cf->dlhandle, "cdcGetAttributeInt");

        cf->cdcFreeString = (void (*)
                            (char *string)
                           ) dlsym(cf->dlhandle, "cdcFreeString");

        cf->cdcFreeObject = (void (*)
                            (cdcObject_t obj)
                            ) dlsym(cf->dlhandle, "cdcFreeObject");

        cf->cdcGetCdcCodeStr = (const char* (*)
                                (cdcCode cc)
                               ) dlsym(cf->dlhandle, "cdcGetCdcCodeStr");

        cf->cdcGetErrSystemStr = (const char* (*)
                                  (cdcErrSystem errSystem)
                                 ) dlsym(cf->dlhandle, "cdcGetErrSystemStr");

        cf->cdcGetLastProtocolError = (cdcCode (*)
                                       (cdcSession_t session,
                                        cdcErrSystem *errSystem,
                                        int *errCode,
                                        const char **errString)
                                      ) dlsym(cf->dlhandle, "cdcGetLastProtocolError");


        ntstatus = NT_STATUS_OK;
    }

ExitRoutine:
    DEBUG(10, ("idmap_cdc_load_lib: libpath=%s ntstatus=0x%x\n",
              libpath, NT_STATUS_V(ntstatus)));

    return ntstatus;
}

/**
 * Map a user ID to a SID.
 * 
 * @param cf        - CAPI session handle and functions
 * @param idmap     - Samba ID Mapping structure,
 *
 * @return NTSTATUS - NT_STATUS_OK if success.
 *                  - See cdcNTStatusMap table for possible errors
 */
static NTSTATUS idmap_cdc_get_sid_from_uid(CapiFuncs *cf, struct id_map *idmap)
{
    cdcObject_t obj = 0;
    char *sidstr = 0;
    cdcCode ccode = CDC_CODE_FAILURE;
    NTSTATUS status = NT_STATUS_OK;

    CDC_EXEC(cdcLookupObjectByUnixId(cf->session, CDC_OBJTYPE_USER, idmap->xid.id, &obj));

    if (ccode != CDC_CODE_SUCCESS)
    {
        DEBUG(10, ("cdcLookupObjectByUnixId failed uid=%d ccode=%d %s\n",
                   idmap->xid.id,
                   ccode,
                   cf->cdcGetCdcCodeStr(ccode)));
        idmap_cdc_err_info(cf, ccode);
        goto ExitRoutine;
    }
    
    CDC_EXEC(cdcGetAttributeStr(cf->session, obj, CDC_ATTR_SID, &sidstr));
    if (ccode != CDC_CODE_SUCCESS)
    {
        DEBUG(2, ("Get User SID attribute failed uid=%d ccode=%d %s\n",
                   idmap->xid.id,
                   ccode,
                   cf->cdcGetCdcCodeStr(ccode)));
        idmap_cdc_err_info(cf, ccode);
        goto ExitRoutine;
    }

    string_to_sid(idmap->sid, sidstr);

ExitRoutine:

    idmap_cdc_obj_info(ccode, "UID", cf, obj, idmap, sidstr);

    cf->cdcFreeString(sidstr);
    cf->cdcFreeObject(obj);

    status = ccode_2_ntstatus(ccode);

    return status;
}

/**
 * Map a group ID to a SID.
 * 
 * @param cf        - CAPI session handle and functions
 * @param idmap     - Samba ID Mapping structure,
 *
 * @return NTSTATUS - NT_STATUS_OK if success.
 *                  - See cdcNTStatusMap table for possible errors
 */
static NTSTATUS idmap_cdc_get_sid_from_gid(CapiFuncs *cf, struct id_map *idmap)
{
    cdcObject_t obj = 0;
    char *sidstr = 0;
    cdcCode ccode = CDC_CODE_FAILURE;
    NTSTATUS ntstatus = NT_STATUS_OK;

    CDC_EXEC(cdcLookupObjectByUnixId(cf->session, CDC_OBJTYPE_GROUP, idmap->xid.id, &obj));

    if (ccode != CDC_CODE_SUCCESS)
    {
        DEBUG(10, ("cdcLookupObjectByUnixId failed gid=%d ccode=%d %s\n",
                   idmap->xid.id,
                   ccode,
                   cf->cdcGetCdcCodeStr(ccode)));
        idmap_cdc_err_info(cf, ccode);
        goto ExitRoutine;
    }

    ccode = cf->cdcGetAttributeStr(cf->session, obj, CDC_ATTR_SID, &sidstr);
    if (ccode != CDC_CODE_SUCCESS)
    {
        DEBUG(2, ("Get Group SID attribute failed uid=%d ccode=%d %s\n",
                  idmap->xid.id,
                  ccode,
                  cf->cdcGetCdcCodeStr(ccode)));
        idmap_cdc_err_info(cf, ccode);
        goto ExitRoutine;
    }

    string_to_sid(idmap->sid, sidstr);

ExitRoutine:

    idmap_cdc_obj_info(ccode, "GID", cf, obj, idmap, sidstr);

    ntstatus = ccode_2_ntstatus(ccode);

    if (obj)
    {
        cf->cdcFreeObject(obj);
    }

    if (sidstr)
    {
        cf->cdcFreeString(sidstr);
    }
    
    return ntstatus;
}

/**
 * Map a Unix ID and object type to a SID.
 * 
 * @param cf        - CAPI session handle and functions
 * @param idmap     - Samba ID Mapping structure,
 *
 * @return NTSTATUS - NT_STATUS_OK if success.
 *                  - See cdcNTStatusMap table for possible errors
 *                  - See cdcNTStatusMap table for possible errors
 */
static NTSTATUS idmap_cdc_get_sid_from_id(CapiFuncs *cf, struct id_map *idmap)
{
    NTSTATUS ntstatus = NT_STATUS_OK;

    if (!cf || !idmap)
    {
        DEBUG(1, ("Error unexpected NULL Parameters: cf=%p idmap=%p\n", cf, idmap));
        ntstatus = NT_STATUS_INVALID_PARAMETER;
        goto ExitRoutine;
    }

    DEBUG(10, ("id=%d id_type=%d\n", idmap->xid.id, idmap->xid.type));

    switch (idmap->xid.type)
    {
        case ID_TYPE_UID:
            ntstatus = idmap_cdc_get_sid_from_uid(cf, idmap);
            break;

        case ID_TYPE_GID:
            ntstatus = idmap_cdc_get_sid_from_gid(cf, idmap);
            break;

        default:
            DEBUG(1, ("Unsupported id_type=%d\n", idmap->xid.type));
            ntstatus = NT_STATUS_INVALID_PARAMETER;
            break;
    }

ExitRoutine:

    if (idmap)
    {
        idmap->status = NT_STATUS_IS_OK(ntstatus) ? ID_MAPPED : ID_UNMAPPED;
    }

    DEBUG(10, ("ntstatus=0x%x\n", NT_STATUS_V(ntstatus)));

    return ntstatus;
}

/**
 * Map a SID to a Unix UID or GID and an Object Type.
 *
 * @param cf        - CAPI session handle and functions
 * @param idmap     - Samba ID Mapping structure,
 *
 * @return NTSTATUS - NT_STATUS_OK if success.
 *                  - See cdcNTStatusMap table for possible errors
 */
static NTSTATUS idmap_cdc_get_id_from_sid(CapiFuncs *cf, struct id_map *idmap)
{
    cdcObject_t obj = 0;
    fstring sidstr;
    cdcCode ccode = CDC_CODE_SUCCESS;

    NTSTATUS ntstatus = NT_STATUS_OK;

    if (!cf || !idmap)
    {
        DEBUG(1, ("Error unexpected NULL Parameters: cf=%p idmap=%p\n", cf, idmap));
        ccode = CDC_CODE_BAD_PARAM;
        goto ExitRoutine;
    }

    sid_to_string(sidstr, idmap->sid);

    DEBUG(10, ("sid=%s\n", sidstr));

    if (strncmp(sidstr, "S-0-0", 5) == 0)
    {
        DEBUG(10, ("Skipping empty or bad sid=%s\n", sidstr));
        ccode = CDC_CODE_BAD_PARAM;
        goto ExitRoutine;
    }

    CDC_EXEC(cdcLookupObjectBySid(cf->session, sidstr, &obj));

    if (ccode != CDC_CODE_SUCCESS)
    {
        DEBUG(10, ("cdcLookupObjectBySid %s failed: ccode=%d %s\n",
                   sidstr,
                   ccode,
                   cf->cdcGetCdcCodeStr(ccode)));
        idmap_cdc_err_info(cf, ccode);
        goto ExitRoutine;
    }

    if (cf->cdcIsObjectCategory(cf->session, obj, CDC_AD_CATEGORY_USER))
    {
        idmap->xid.type = ID_TYPE_UID;

        CDC_EXEC(cdcGetAttributeInt(cf->session, obj, CDC_ATTR_UID, &idmap->xid.id));
        if (ccode != CDC_CODE_SUCCESS)
        {
            DEBUG(7, ("cdcGetAttributeInt(uid) failed: ccode=%d %s",
                      ccode,
                      cf->cdcGetCdcCodeStr(ccode)));
            idmap_cdc_err_info(cf, ccode);
        }
    }
    else if (cf->cdcIsObjectCategory(cf->session, obj, CDC_AD_CATEGORY_GROUP))
    {
        idmap->xid.type = ID_TYPE_GID;
        CDC_EXEC(cdcGetAttributeInt(cf->session, obj, CDC_ATTR_GID, &idmap->xid.id));
        if (ccode != CDC_CODE_SUCCESS)
        {
            DEBUG(3, ("cdcGetAttributeInt(gid) failed: ccode=%d %s",
                      ccode,
                      cf->cdcGetCdcCodeStr(ccode)));
            idmap_cdc_err_info(cf, ccode);
        }
    }
    else
    {
        DEBUG(10, ("Object is not a user or a group\n"));
        ccode = CDC_CODE_NO_SUCH_OBJECT;
    }

ExitRoutine:
    ntstatus = ccode_2_ntstatus(ccode);

    if (idmap)
    {
        idmap->status = NT_STATUS_IS_OK(ntstatus) ? ID_MAPPED : ID_UNMAPPED;
    }

    idmap_cdc_obj_info(ccode, "SID", cf, obj, idmap, sidstr);
    
    if (obj)
    {
        cf->cdcFreeObject(obj);
    }

    DEBUG(10, ("ntstatus=0x%x\n", NT_STATUS_V(ntstatus)));

    return ntstatus;
}

/**
 * IDMAP API close function
 * Close out the CAPI session and unlink from the dynamically loaded
 * CAPI library.
 *
 * @return NTSTATUS - Always returns NT_STATUS_OK.
 */
static NTSTATUS idmap_cdc_close(struct idmap_domain *dom)
{
    DEBUG(3, ("idmap_cdc_close\n"));

    if (dom)
    {
        CapiFuncs *cf = (CapiFuncs*)dom->private_data;

        if (cf)
        {
            if (cf->session)
            {
                cf->cdcSessionClose(cf->session);
            }

            if (cf->dlhandle)
            {
                dlclose(cf->dlhandle);
            }

            memset(cf, 0, sizeof(*cf));

            TALLOC_FREE(dom->private_data);
        }

        dom->private_data = 0;
        dom->initialized = False;
    }

    return NT_STATUS_OK;
}

/**
 * idmap API init function
 *
 * Try to dynamically load the CAPI library and open a CAPI session.
 *
 * @param capi_lib_path - Path to CAPI library as defined in smb.conf.
 */
static NTSTATUS idmap_cdc_initialize(struct idmap_domain *dom)
{
    CapiFuncs *cf = 0;
    cdcCode ccode = CDC_CODE_FAILURE;
    NTSTATUS ntstatus = NT_STATUS_OK;

    DEBUG(3, ("params=%s initialized=%s\n",
              dom->params ? dom->params : "<NULL>",
              dom->initialized ? "True" : "False"));

    if (dom->initialized == True)
    {
        goto ExitRoutine;
    }

    if ((cf = TALLOC_ZERO_P(dom, CapiFuncs)) == NULL)
    {
        DEBUG(0, ("Out of memory!\n"));
        ntstatus = NT_STATUS_NO_MEMORY;
        goto ExitRoutine;
    }

    dom->private_data = cf;
    
    ntstatus = idmap_cdc_load_lib(dom);
    if (NT_STATUS_IS_ERR(ntstatus))
    {
        goto ExitRoutine;
    }
    
    ccode = cf->cdcSessionOpen(CDC_CAPI_MAJOR_VERSION,
                               CDC_CAPI_MINOR_VERSION,
                               &cf->session);
    if (ccode != CDC_CODE_SUCCESS)
    {
        DEBUG(1, ("cdcSessionOpen failed. DirectControl not loaded? ccode=%d %s\n",
                  ccode,
                  cf->cdcGetCdcCodeStr(ccode)));

        ntstatus = ccode_2_ntstatus(ccode);
        goto ExitRoutine;
    }
    else
    {
        // Limit ourselves to just zoned objects.
        unsigned int writeVal = 1;
        size_t valSize = sizeof(unsigned int);
        ccode = cf->cdcSetOption(cf->session, CDC_OPT_UNIX_ONLY, (void *)&writeVal, valSize);
        if (ccode != CDC_CODE_SUCCESS)
        {
            DEBUG(1, ("cdcSetOption (Unix only) failed: Proceeding anyway. ccode=%d %s\n",
                      ccode,
                      cf->cdcGetCdcCodeStr(ccode)));
            idmap_cdc_err_info(cf, ccode);
        }
    }

    dom->initialized = True;

ExitRoutine:
    if (NT_STATUS_IS_ERR(ntstatus))
    {
        idmap_cdc_close(dom);
    }
    
    DEBUG(3, ("ntstatus=0x%x\n", NT_STATUS_V(ntstatus)));
    return ntstatus;
}

/**
 * Sanity check the input parameters to our major mapping functions.
 *
 * @param msg - Indicates where we were called from in log messages.
 * @param dom - Domain Specific structure. Contains CAPI session handle.
 * @param ids - An array of pointers to ID mapping structures.
 *
 * @return ntstatus - NT_STATUS_OK structures are okay.
 */
static NTSTATUS check_params(const char *msg, struct idmap_domain *dom, struct id_map **ids)
{
    NTSTATUS ntstatus = NT_STATUS_OK;

    if (dom == 0 || ids == 0)
    {
        DEBUG(1, ("%s: Error unexpected NULL Parameters: dom=%p ids=%p\n",
                  msg, dom, ids));
        ntstatus = NT_STATUS_INVALID_PARAMETER;
        goto ExitRoutine;
    }

    if (dom->initialized == False)
    {
        DEBUG(5, ("%s: Initializing dom...\n", msg));

        ntstatus = idmap_cdc_initialize(dom);

        if (NT_STATUS_IS_ERR(ntstatus))
        {
            goto ExitRoutine;
        }
    }

    if (!dom->private_data)
    {
        DEBUG(1, ("%s: Internal CAPI session structure is missing!\n", msg));
        ntstatus = NT_STATUS_INVALID_PARAMETER;
        goto ExitRoutine;
    }

ExitRoutine:
    return ntstatus;
}

/**
 * Map a list of SIDs to Unix IDs.
 *
 * @param dom - Domain Specific structure. Contains CAPI session handle.
 * @param ids - An array of pointers to ID mapping structures.
 *
 * @return ntstatus - NT_STATUS_OK if no major system errors.
 */
static NTSTATUS idmap_cdc_sids_to_unixids(struct idmap_domain *dom, struct id_map **ids)
{
    NTSTATUS ntstatus = check_params("SIDs to IDs", dom, ids);

    if (NT_STATUS_IS_OK(ntstatus))
    {
        CapiFuncs *cf = (CapiFuncs*)dom->private_data;
        int i;

        for (i=0; ids[i]; i++)
        {
            ntstatus = idmap_cdc_get_id_from_sid(cf, ids[i]);

            if ((!NT_STATUS_IS_OK(ntstatus)) && (!NT_STATUS_EQUAL(ntstatus, NT_STATUS_NONE_MAPPED)))
            {
                /* Some fatal error occurred, log it */
                DEBUG(3, ("Unexpected error resolving a SID: ntstatus=0x%x\n",
                          NT_STATUS_V(ntstatus)));
            }

            ntstatus = NT_STATUS_OK;
        }
    }

    return ntstatus;
}

/**
 * Map a list of Unix Ids to SIDs.
 *
 * @param dom - Domain Specific structure. Contains CAPI session handle.
 * @param ids - An array of pointers to ID mapping structures.
 *
 * @return ntstatus - NT_STATUS_OK if no major system errors.
 */
static NTSTATUS idmap_cdc_unixids_to_sids(struct idmap_domain *dom, struct id_map **ids)
{
    NTSTATUS ntstatus = check_params("IDs to SIDs", dom, ids);

    if (NT_STATUS_IS_OK(ntstatus))
    {
        CapiFuncs *cf = (CapiFuncs*)dom->private_data;
        int i;

        for (i=0; ids[i]; i++)
        {
            ntstatus = idmap_cdc_get_sid_from_id(cf, ids[i]);
            if ((!NT_STATUS_IS_OK(ntstatus)) &&( !NT_STATUS_EQUAL(ntstatus, NT_STATUS_NONE_MAPPED)))
            {
                /* Some fatal error occurred, log it */
                DEBUG(3, ("Unexpected error resolving an ID (%d) ntstatus=0x%x\n",
                          ids[i]->xid.id,
                          NT_STATUS_V(ntstatus)));
            }

            ntstatus = NT_STATUS_OK;
        }
    }

    return ntstatus;
}

#define STRUCTURE_VERSION_NUMBER 2

/*
 * Version information structure returned in response to
 * idmap_cdc_get_version_info.
 */
typedef struct
{
    const char *samba_version_string;
    const char *idmap_version;
    int idmap_interface_version;
    int cdc_major_version;
    int cdc_minor_version;
} IDMapCdcVersionInfo;

/*
 * This is an IDMAP version 4 style table (For Samba 3.0.25)
 */
static struct idmap_methods cdc_methods = {
    .init = idmap_cdc_initialize,
    .unixids_to_sids = idmap_cdc_unixids_to_sids,
    .sids_to_unixids = idmap_cdc_sids_to_unixids,
    .close_fn = idmap_cdc_close
};

/**
 * Called by Winbind to get us to register our dispatch table.
 *
 * If this idmapper is built as a shared object library, its name
 * "idmap_cdc_init" will be translated.
 * Samba's build-time generated config.h file will contain:
 * 
 * #define idmap_cdc_init init_module
 *
 * This allows its shared object loader code to always bind to
 * init_module.
 *
 * @return - NT Status code: NT_STATUS_OK for success.
 */
NTSTATUS idmap_cdc_init(void)
{
    NTSTATUS ntstatus =  smb_register_idmap(SMB_IDMAP_INTERFACE_VERSION,
                                           "cdc",
                                           &cdc_methods);
    
    DEBUG(3, ("init_module (cdc): ntstatus 0x%x\n", NT_STATUS_V(ntstatus)));

    return ntstatus;
}


/**
 * CDC specific function that returns idmap version information.
 *
 * @param version_info  - Place to return version structure pointer.
 * @return              - The internal structure version.
 */
int idmap_cdc_get_version_info(void **version_info)
{
    static IDMapCdcVersionInfo cdc_version_info;

    if (version_info)
    {
        cdc_version_info.samba_version_string = SAMBA_VERSION_STRING;
        cdc_version_info.idmap_interface_version = SMB_IDMAP_INTERFACE_VERSION;
        cdc_version_info.idmap_version = SAMBA_VERSION_VENDOR_SUFFIX;
        cdc_version_info.cdc_major_version = CDC_CAPI_MAJOR_VERSION;
        cdc_version_info.cdc_minor_version = CDC_CAPI_MINOR_VERSION;

        *version_info = (void *)&cdc_version_info;
    }

    DEBUG(3, ("samba=%s idmap=%s idmap_interface=%d major=%d minor=%d\n",
              cdc_version_info.samba_version_string,
              cdc_version_info.idmap_version,
              cdc_version_info.idmap_interface_version,
              cdc_version_info.cdc_major_version,
              cdc_version_info.cdc_minor_version));

    return STRUCTURE_VERSION_NUMBER;
}
