/* 
   Unix SMB/CIFS implementation.

   trivial database library

   Copyright (C) Andrew Tridgell 1999-2004
   Copyright (C) David Daugherty 2005

     ** NOTE! The following LGPL license applies to the libtdb
     ** library. This does NOT imply that all of Samba is released
     ** under the LGPL

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#define STANDALONE
#include "config.h"

#include <stdio.h>

#ifdef HAVE_FCNTL_H
#include <fcntl.h>
#else
#ifdef HAVE_SYS_FCNTL_H
#include <sys/fcntl.h>
#endif
#endif

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#include <signal.h>
#include <sys/types.h>
#include <sys/stat.h>


#ifdef HAVE_STRING_H
#include <string.h>
#endif

#ifdef TIME_WITH_SYS_TIME
#include <sys/time.h>
#include <time.h>
#else
#ifdef HAVE_SYS_TIME_H
#include <sys/time.h>
#else
#include <time.h>
#endif
#endif


#ifdef HAVE_MEMORY_H
#include <memory.h>
#endif

#ifdef HAVE_MALLOC_H
#include <malloc.h>
#endif

#include "tdb.h"
#include "libtdb.h"

#ifndef EMPTY_THING
#define EMPTY_THING(s) (!s || !*s)
#endif

#ifndef MAX_TDB_KEY_SIZE
#define MAX_TDB_KEY_SIZE 128  /* Arbitrary size */
#endif

/**
 * get8
 *
 * Get an unsigned byte from the buffer. Building block for the rest of the functions.
 */
static uint8 get8(const uint8 *p)
{
    return *p;
}

/**
 * put8
 *
 * Put an unsigned byte into buffer. Building block for the rest of the functions.
 */
static void put8(uint8 *p, uint8 val)
{
    *p = val;
}

/**
 * getlohi32
 *
 * Get an unsigned 32 bit value from a buffer where it is
 * stored in little endian format and return it in host order.
  */
static uint32 getlohi32(const uint8 *p)
{
    return get8(p++) + (get8(p++) << 8) + (get8(p++) << 16) + (get8(p) << 24);
}

/**
 * putlohi32
 *
 * Put a host-ordered unsigned 32 bit value into a buffer in little endian format
 */
static void putlohi32(uint8 *p, uint32 val)
{
    put8(p++, (uint8)val);
    put8(p++, (uint8)(val >> 8));
    put8(p++, (uint8)(val >> 16));
    put8(p++, (uint8)(val >> 24));
}

/**
 * Set the machine password info in secrets.tdb and update the last password change time.
 *
 * @param tdbfile - full path of the TDB file to set the password in
 * @param domain - the NTLM domain name that we are changing the machine password for
 * @param password - the new machine account password
 * @param sec_channel - the security channel setting. See get_default_sec_channel(void) for examples
 *
 * @return - LIBTDB return code
 */
int libtdb_set_machineinfo(const char *tdbfile, const char *domain, const char *password, uint32 sec_channel)
{
    char key[MAX_TDB_KEY_SIZE];
    TDB_HANDLE tdb_handle;
    time_t now = time(0);
    int ltdbcode = LIBTDB_BAD_PARAM;

    uint8 lohi_value[8];
    
    if (EMPTY_THING(tdbfile) || EMPTY_THING(domain) || EMPTY_THING(password))
        goto EXIT2;
    
    if ((ltdbcode = libtdb_open(tdbfile, O_RDWR|O_CREAT, 0600, &tdb_handle)) != LIBTDB_SUCCESS)
        goto EXIT2;

    snprintf(key, MAX_TDB_KEY_SIZE, "SECRETS/MACHINE_PASSWORD/%s", domain);
    if ((ltdbcode = libtdb_write(tdb_handle, key, CONST_CAST(void*, password), strlen(password)+1)) != LIBTDB_SUCCESS)
        goto EXIT1;

    /* Really it needs to be stored little endian */
    putlohi32(lohi_value, now);
    snprintf(key, MAX_TDB_KEY_SIZE, "SECRETS/MACHINE_LAST_CHANGE_TIME/%s", domain);
    if ((ltdbcode = libtdb_write(tdb_handle, key, (void*)lohi_value, sizeof(now))) != LIBTDB_SUCCESS)
        goto EXIT1;

    /* Really it needs to be stored little endian */
    putlohi32(lohi_value, sec_channel);
    snprintf(key, MAX_TDB_KEY_SIZE, "SECRETS/MACHINE_SEC_CHANNEL_TYPE/%s", domain);
    ltdbcode = libtdb_write(tdb_handle, key, (void*)lohi_value, sizeof(sec_channel));

EXIT1:
    libtdb_close(tdb_handle);

EXIT2:
    return ltdbcode;
}

/**
 * Get the machine password info from secrets.tdb and update the last password change time.
 *
 * @param tdbfile - full path of the TDB file to set the password in
 * @param domain - the NTLM domain name for the information
 * @param password - Place to return the passowrd in
 * @param size - Size of the password buffer
 * @param real_size - The actual size of the password 
 * @param last_pwd_change_time - Time password was last changed
 * @param sec_channel - Place for the returned security channel setting
 *
 * @return - LIBTDB return code
 */
int libtdb_get_machineinfo(const char *tdbfile, const char *domain, char *password, size_t size, size_t *real_size,
                           time_t *last_pwd_change_time, uint32 *sec_channel)
{
    char key[MAX_TDB_KEY_SIZE];
    TDB_HANDLE tdb_handle;
    uint8 lohi_value[8];

    int ltdbcode = LIBTDB_BAD_PARAM;

    if (EMPTY_THING(tdbfile) || EMPTY_THING(domain) || !password)
        goto EXIT2;

    if ((ltdbcode = libtdb_open(tdbfile, O_RDONLY, 0600, &tdb_handle)) != LIBTDB_SUCCESS)
        goto EXIT2;

    snprintf(key, MAX_TDB_KEY_SIZE, "SECRETS/MACHINE_PASSWORD/%s", domain);
    if ((ltdbcode = libtdb_read(tdb_handle, key, password, size, real_size)) != LIBTDB_SUCCESS)
        goto EXIT1;

    snprintf(key, MAX_TDB_KEY_SIZE, "SECRETS/MACHINE_LAST_CHANGE_TIME/%s", domain);
    if ((ltdbcode = libtdb_read(tdb_handle, key,  lohi_value,
                                sizeof(*last_pwd_change_time), 0)) != LIBTDB_SUCCESS)
        goto EXIT1;

    *last_pwd_change_time = getlohi32(lohi_value);
    snprintf(key, MAX_TDB_KEY_SIZE, "SECRETS/MACHINE_SEC_CHANNEL_TYPE/%s", domain);
    ltdbcode = libtdb_read(tdb_handle, key, lohi_value, sizeof(*sec_channel), 0);

    *sec_channel= getlohi32(lohi_value);
    
EXIT1:
    libtdb_close(tdb_handle);

EXIT2:
    return ltdbcode;
}

/**
 * Open a TDB File
 *
 * @param tdbfile - full path of the TDB file to set the password in
 * @param open_flags - see fcntl.h
 * @param mode - file permissions to set if file was created.
 * @param tdb_handle - place to return the TDB_CONTEXT.
 *
 * @return - LIBTDB return code
 */
int libtdb_open(const char *tdbfile, int open_flags, mode_t mode, TDB_HANDLE *tdb_handle)
{
    TDB_CONTEXT *tdb = tdb_open(tdbfile, 0, TDB_DEFAULT, open_flags, mode);
    int ltdbcode = LIBTDB_SUCCESS;

    *tdb_handle = (TDB_HANDLE)tdb;
    
    if (tdb == NULL)
        ltdbcode = LIBTDB_OPEN_FAILED;

    return ltdbcode;
}
        
int libtdb_close(TDB_HANDLE tdb_handle)
{
    int ltdbcode = LIBTDB_CLOSE_FAILED;
    
    if (tdb_handle)
        if (tdb_close((TDB_CONTEXT *)tdb_handle))
            ltdbcode = LIBTDB_SUCCESS;

    return ltdbcode;
}
    
/**
 * Write a TDB record
 *
 * @param tdb_handle - opened TDB context
 * @param key - the record key
 * @param data - the record data
 * @param size - the record key
 *
 * @return - LIBTDB return code
 */
int libtdb_write(TDB_HANDLE tdb_handle, const char *key, const void *data, size_t size)
{
    TDB_DATA k;
    TDB_DATA d;
    int ltdbcode = LIBTDB_SUCCESS;

    k.dptr = CONST_CAST(char*, key);
    k.dsize = strlen(key);

    d.dptr = CONST_CAST(char*, data);
    d.dsize = size;

    if (tdb_store((TDB_CONTEXT *)tdb_handle, k, d, TDB_REPLACE) != 0)
        ltdbcode = LIBTDB_WRITE_FAILED;

    return ltdbcode;
}

/**
 * Read a TDB record
 *
 * @param tdb_handle - opened TDB context
 * @param key - the record key
 * @param data - buffer to store the read data
 * @param size - max size of supplied data buffer
 * @param real_size - place to store the actual size of the read data.
 *
 * @return - LIBTDB return code
 */
int libtdb_read(TDB_HANDLE tdb_handle, const char * key, void *data, size_t size, size_t *real_size)
{
    TDB_DATA k;
    TDB_DATA d;
    int ltdbcode = LIBTDB_SUCCESS;
    
    k.dptr = CONST_CAST(char*, key);
    k.dsize = strlen(key);

    d = tdb_fetch(tdb_handle, k);

    if (real_size)
        *real_size = d.dsize;

    if (d.dsize == 0)
        ltdbcode = LIBTDB_NO_DATA;
    else if (d.dsize > size)
        ltdbcode = LIBTDB_BUF_TOO_SMALL;
    else
        memcpy(data, d.dptr, d.dsize);

    if (d.dptr)
        free(d.dptr);
    
    return ltdbcode;
}

/**
 * This function gets called for each record found in the TDB that we are traversing
 * @param tc - Open TDB context
 * @param key - The key of the current record
 * @param data - The data of the current record
 * @param state - the client's state parameter - in this case a pointer to a traverse_control structure.
 *
 * @return - LIBTDB return code
 */
static int libtdb_traverse_callback(TDB_CONTEXT *tc, TDB_DATA key, TDB_DATA data, void *state)
{
    traverse_control *tctl = (traverse_control*)state;

    tctl->handle = (TDB_HANDLE)tc;
    tctl->key = key.dptr;
    tctl->keysize = key.dsize;
    tctl->data = (void *)data.dptr;
    tctl->datasize = data.dsize;

    return tctl->callback(tctl);
}

/**
 * Traverse a Samba database file, reading each record sequentially.
 *
 * @param tdbfile - the TDB file to traverse
 * @param tctl - Control structure for the traverse
 * @param count - Place to store the number of records read.
 * @return ltdcode - See libtdb.h for codes
 */
int libtdb_traverse(const char *tdbfile, traverse_control *tctl)
{
    TDB_HANDLE tdb_handle;
    int ltdbcode = LIBTDB_BAD_PARAM;
    
    if (EMPTY_THING(tdbfile) || !tctl)
        goto EXIT;

    if ((ltdbcode = libtdb_open(tdbfile,  O_RDONLY, 0600, &tdb_handle)) != LIBTDB_SUCCESS)
        goto EXIT;

    tctl->count = tdb_traverse((TDB_CONTEXT *)tdb_handle, libtdb_traverse_callback, (void*)tctl);

    if (tctl->count == -1)
        ltdbcode = LIBTDB_TRAVERSE_FAILED;

    libtdb_close(tdb_handle);

EXIT:
    return ltdbcode;
}
