/* 
   IDMap Unit Test Functions
   Copyright (C) David Daugherty 2007

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 or version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "includes.h"

#define MAX_IDS 16

const char *domainName = "ALLDOMAINS";

typedef struct
{
    void *dlhandle;
    NTSTATUS (*init_module)(void);
    int (*version_info)(void **vinfo);
} idmap_funcs;


struct idmap_function_entry {
    const char *name;
    struct idmap_methods *methods;
    struct idmap_function_entry *prev,*next;
};


struct idmap_function_entry ife = {0};

extern NTSTATUS init_module(void);

/*
 * Version information structure returned in response to
 * idmap_cdc_get_version_info.
 */
typedef struct
{
    const char *samba_version_string;
    const char  *idmap_version;
    int idmap_interface_version;
    int cdc_major_version;
    int cdc_minor_version;
} IDMapCdcVersionInfo;

static void print_version_info(idmap_funcs *idm)
{
    IDMapCdcVersionInfo *vinfo;
    int struct_version = idm->version_info((void**)&vinfo);
    if (struct_version == 2)
    {
        printf("Samba Version:      %s\n", vinfo->samba_version_string);
        printf("IDMap Version:      %s\n", vinfo->idmap_version);
        printf("IDMap Interface Version:      %d\n", vinfo->idmap_interface_version);		
        printf("CDC Major Version:  %d\n", vinfo->cdc_major_version);
        printf("CDC Minor Version:  %d\n", vinfo->cdc_minor_version);
    }
    else
    {
        printf("Unknown structure version: %d\n", struct_version);
    }
}
    
static NTSTATUS idm_load(const char *libpath, idmap_funcs *idm)
{
    NTSTATUS ntstatus = NT_STATUS_NO_SUCH_FILE;

    memset(idm, 0, sizeof(*idm));

    idm->dlhandle = dlopen(libpath, RTLD_LAZY);

    if (idm->dlhandle != NULL)
    {
        idm->init_module = (NTSTATUS (*) (void)
                           ) dlsym(idm->dlhandle, "init_module");

        idm->version_info = (int (*)(void **vinfo)
                            ) dlsym(idm->dlhandle, "idmap_cdc_get_version_info");

        ntstatus = NT_STATUS_OK;
    }

    return ntstatus;
}

static void idm_unload(idmap_funcs *idm)
{
    if (idm && idm->dlhandle)
    {
        dlclose(idm->dlhandle);
        memset(idm, 0, sizeof(*idm));
    }
}

static void display_help_and_exit(pstring libpath)
{
    printf("testidmap: Unit test program for Samba's (CDC) Idmapper\n");
    printf("Before it can do anything, this test program must first\n");
    printf("dynamically load and bind to the cdc idmapper library\n");
    printf("\n");
    printf("Usage:  testidmap -l <cdc library path> <options>\n");
    printf("Examples:\n");
    printf("\n");
    printf("testidmap -l %s -v\n",libpath);
    printf("    Calls internal idmap_cdc_get_version_info function\n");
    printf("    to get cdc.so version information and prints it out.\n");
    printf("\n");
    printf("testidmap -l %s -u -i 10000 -i 10001 -i 10002\n",libpath);
    printf("    looks up users by ids, 10000, 10001, and 10002 and displays their sids\n");
    printf("    then it looks up the same users by the sids returned and displays their ids.\n");
    printf("\n");            
    printf("    testidmap -l %s -g -i 10000 -i 10001 -i 10002\n",libpath);
    printf("    looks up groups by ids, 10000, 10001, and 10002 and displays their sids\n");
    printf("    then it looks up the same groups by the sids returned and displays their ids.\n");
    printf("\n");

    exit(55);
}

void dump_idmaps(struct id_map *pidmaps[])
{
    int i;

    for (i=0; pidmaps[i]; ++i)
    {
        struct id_map* idmap = pidmaps[i];
        fstring sidstr;

        sid_to_string(sidstr, idmap->sid);
        printf("type:%s id:%d %s\n", idmap->xid.type == 1 ? "UID" : "GID", idmap->xid.id, sidstr);
    }
}

int main(int argc, const char *argv[])
{
    struct idmap_domain *dom = 0;

    struct id_map *pidmaps[MAX_IDS];
    struct id_map idmaps[MAX_IDS];
    DOM_SID sids[MAX_IDS];
    int idx = 0;
    
    int opt;

    idmap_funcs idm = {0};

    NTSTATUS ntstatus;

    enum id_type idtype = ID_TYPE_UID;

    int vinfo = 0;
    poptContext pc;

    struct poptOption long_options[] = {
        POPT_AUTOHELP
        {"help",    'h', POPT_ARG_NONE,     0,      'h',      "Help\n"},
        {"version", 'v', POPT_ARG_NONE,     0,      'v',      "Version Info\n"},
        {"user",    'u', POPT_ARG_NONE,     0,      'u',      "UID"},
        {"group",   'g', POPT_ARG_NONE,     0,      'g',      "GID"},
        {"id",      'i', POPT_ARG_STRING,   0,      'i',      "Object ID"},
        {"lib",     'l', POPT_ARG_STRING,   0,      'l',      "ID Mapper Library\n"},
        POPT_TABLEEND
    };
    pstring libpath;

    pstr_sprintf(libpath, "%s/idmap/cdc.%s", dyn_LIBDIR, dyn_SHLIBEXT);
    memset(pidmaps, 0, sizeof(pidmaps));
    memset(idmaps, 0, sizeof(idmaps));
    memset(sids, 0, sizeof(sids));

    
    load_case_tables();

    if (!lp_load(dyn_CONFIGFILE, True, False, False, True))
    {
        printf("Unable to load config file: %s\n", dyn_CONFIGFILE);
        goto ExitRoutine;
    }
    

    pc = poptGetContext(NULL, argc, argv, long_options,
                        POPT_CONTEXT_KEEP_FIRST);

    poptSetOtherOptionHelp(pc, "Unknown option");

    while ((opt = poptGetNextOpt(pc)) != -1)
    {
        switch (opt)
        {
            case 'u':
                idtype = ID_TYPE_UID;
                break;

            case 'g':
                idtype = ID_TYPE_GID;
                break;

            case 'i':
                idmaps[idx].xid.id = atoi(poptGetOptArg(pc));
                idmaps[idx].xid.type = idtype;
                idmaps[idx].sid = &sids[idx];
                pidmaps[idx] = &idmaps[idx];
                idx++;
                break;

            case 'v':
                vinfo = 1;
                break;

            case 'l':
                pstrcpy(libpath, poptGetOptArg(pc));
                break;

            case 'h':
                display_help_and_exit(libpath);
                break;
                
            default:
                printf("\nInvalid option %s: %s\n", 
                       poptBadOption(pc, 0), poptStrerror(opt));
                display_help_and_exit(libpath);
                break;
        }
    }

    if (!libpath)
    {
        printf("libpath is a required parameter\n");

        ntstatus = NT_STATUS_INVALID_PARAMETER;
        goto ExitRoutine;
    }

    ntstatus = idm_load(libpath, &idm);

    if (NT_STATUS_IS_ERR(ntstatus))
    {
        printf("Failed to load idmapper: %s\n", libpath);
        printf("Check the path and try again\n");

        goto ExitRoutine;
    }

    if (vinfo)
    {
        print_version_info(&idm);
        goto ExitRoutine;
    }

    ntstatus = idm.init_module();

    if (NT_STATUS_IS_ERR(ntstatus))
    {
        printf("cdc 'init_module' (a.k.a idmap_cdc_initialize) function failed.\n");
        printf("Check samba.3.0.26a/source/include/config.h and look for the line\n");
        printf("\n");
        printf("    define idmap_cdc_initialize init_module\n");
        printf("\n");
        printf("If its not there, try deleting the samba build directory and re-running\n");
        printf("build.sh samba\n");
        
        goto ExitRoutine;
    }

    dom = TALLOC_ZERO_P(NULL, struct idmap_domain);
    dom->name = domainName;

    if (lp_idmap_backend())
    {
        const char **compat_list = lp_idmap_backend();
        const char *compat_backend =  *compat_list;
        const char *p = strchr(compat_backend, ':');
        if (p != NULL)
        {
            dom->params = p+1;
        }
        else
        {
            printf("SMB.CONF compatibilty backend key must include path to CAPI library\n");
            printf("Example: idmap backend = cdc:/usr/share/centrifydc/lib/libcapi.so\n");
            ntstatus = NT_STATUS_INVALID_PARAMETER;
            goto ExitRoutine;
        }
    }
    
    ntstatus = ife.methods->init(dom);

    if (NT_STATUS_IS_ERR(ntstatus))
    {
        printf("CDC Init method failed.\n"); 
        goto ExitRoutine;
    }

    ntstatus = ife.methods->unixids_to_sids(dom, pidmaps);

    if (NT_STATUS_IS_ERR(ntstatus))
    {
        printf("CDC unixids_to_sids method failed.\n"); 
        goto ExitRoutine;
    }


    ntstatus = ife.methods->sids_to_unixids(dom, pidmaps);
    if (NT_STATUS_IS_ERR(ntstatus))
    {
        printf("CDC sids_to_unixids method failed.\n"); 
        goto ExitRoutine;
    }

    dump_idmaps(pidmaps);

ExitRoutine:

    if (ife.methods && ife.methods->close_fn)
    {
        ife.methods->close_fn(dom);
    }

    idm_unload(&idm);

    if (dom)
    {
        talloc_free(dom);
    }

    if (NT_STATUS_IS_ERR(ntstatus))
    {
        printf("Failed: NTStatus = %d (0x%x)\n", NT_STATUS_V(ntstatus), NT_STATUS_V(ntstatus));
    }
    else
    {
        printf("Success\n");
    }
    
    return NT_STATUS_V(ntstatus);
}

NTSTATUS smb_register_idmap(int version, const char *name, struct idmap_methods *methods)
{
    printf ("smb_register_idmap: version=%d, name=%s\n", version, name);

    ife.name = name;
    ife.methods = methods;

    return NT_STATUS_OK;
}

char *ldap_err2string(int rc)
{
    static fstring errstr;
    
    snprintf(errstr, 256, "ldap_err2String: %d\n", rc);

    return (char*)errstr;
}
